from plotly.graph_objects import FigureWidget
import json, os, math
import qgrid
from ipywidgets import FloatRangeSlider, IntRangeSlider, HBox, VBox, Box, Tab, Layout, Output, Dropdown, FloatLogSlider, BoundedFloatText, BoundedIntText
from IPython.display import Javascript, clear_output
import pandas as pd
import nanohub.remote as nr
import types

def getRapptureSchema(self, toolname, force=True, **kwargs):
    timeout=kwargs.get("timeout", 30)      
    self.validateSession()
    if (self.authenticated == False):
        raise ConnectionError('not connected')
    if (self.cached_schema is not None and self.cached_toolname == toolname and force == False):
        return self.cached_schema
    else:
        turl = self.url
        self.url = 'http://instanton2.nanohub.org:8088'
        request = self.requestGet('archive/'+kwargs['squid'], data={'output':'xml'}, timeout=timeout)
        self.url = turl
        rappturexml = request.text
        self.cached_schema = rappturexml
        self.cached_toolname = toolname
    return rappturexml;
    

class PNToyExplorer ():
    URL = r'https://nanohub.org/api'
    TOOL = "pntoy"
    FILTERS =  [
    # 'input.Na',
    # 'input.Nd',
     'input.n_len',
     'input.i_len',
     'input.p_len',
    #'input.i_node',
    #'input.n_node',
    #'input.p_node',
     'input.vn_step',
    # 'input.materialp',
     'input.temperature',
    # 'input.vsweep_high',
    # 'input.impuritylevel',
    # 'input.impuritydoping'
    # 'input.taun',
    # 'input.taup',
    ]
    VALID_RUNS = 1
    REVISION = 0
    RANDOM = 0
    
    def __init__(self, credentials, **kwargs):
        self.session = nr.Tools(credentials)
        self.window = VBox([])
        self.dataframe = None;
        self.active = None
        theme = 'plotly_white' #'plotly_dark'
        layout0 = {
            'template' : theme,
            'margin' : {'r' : 25, 'l' : 25, 't' : 40,'b' : 50 },
            'showlegend' : False,
            'title' : "Energy Band Diagram"
        }
        layout1 = {
            'template' : theme,
            'margin' : {'r' : 25, 'l' : 25, 't' : 40,'b' : 50 },
            'showlegend' : False,
            'title' : "I/C-V Characteristics",
            'yaxis':{
                'side' : 'left',
                'title' : 'Current [A/cm2]',
            },
            'yaxis2':{
                'side' : 'right',
                'title' : 'Capacitance [F/cm2]',
                'overlaying' : 'y',
            },

        }        
        self.plotly_fig = [
            FigureWidget([], layout0),
            FigureWidget([], layout1) 
        ]
            
        self.materialp = Dropdown(options=["Si"], value="Si", description='Material:')
        self.materialp.observe(lambda w:self.updateMain(), names='value')
        self.vsweep_high = BoundedFloatText(
            value=0.6,
            min=0,
            max=2.0,
            step=0.1,
            description='Applied Voltage:',
            disabled=False,
            layout = Layout(width='98%')
        )        
        self.vsweep_high.observe(lambda w:self.updateMain(), names='value')

        self.samples = BoundedIntText(
            value=10,
            min=2,
            max=50,
            step=1,
            description='Samples:',
            disabled=False,
            layout = Layout(width='98%')
        )
        self.samples.observe(lambda w:self.update_filters(), names='value')
        
        self.Na = FloatLogSlider(
            value=2e15,
            base=10,
            min=5,
            max=30,
            step=0.2,
            description='Na',
            continuous_update = False,
            layout = Layout(width='98%')
        )
        self.Na.observe(lambda w:self.updateMain(), names='value')

        self.Nd = FloatLogSlider(
            value=1e15,
            base=10,
            min=5,
            max=30,
            step=0.2,
            description='Nd',
            continuous_update = False,
            layout = Layout(width='98%')
        )
        self.Nd.observe(lambda w:self.updateMain(), names='value')

        self.taun = FloatLogSlider(
            value=1e-10,
            base=10,
            min=-15,
            max=-5,
            step=0.5,
            description='taun',
            continuous_update = False,
            layout = Layout(width='98%')
        )
        self.taun.observe(lambda w:self.updateMain(), names='value')

        self.taup = FloatLogSlider(
            value=1e-10,
            base=10,
            min=-15,
            max=-5,
            step=0.5,
            description='taup',
            continuous_update = False,
            layout = Layout(width='98%')
        )
        self.taup.observe(lambda w:self.updateMain(), names='value')

        layout2 = {
            'template' : theme,
            'margin' : { 'r' : 30, 'l' : 30,'t' : 20, 'b' : 0 },
            'showlegend' : False,
            'yaxis':{
                'range' : [0,1],
                'visible' : False,
                'autorange' : False,
            },            
            'xaxis':{
                'range' : [0,1],
                'visible' : True,
                'autorange' : False,
                'tickangle' : 80
            },
        }
        layout3 = {
            'template' : theme,
            'margin' : {
                'r' : 30,
                'l' : 30,
                't' : 10,
                'b' : 10,
            },
            'showlegend' : False,
            'yaxis':{
                'range' : [0,1],
                'visible' : False,
                'autorange' : False,
            },            
            'xaxis':{
                'range' : [0,1],
                'visible' : False,
                'autorange' : False,
            },
        }
        self.plotly_map = [
            FigureWidget([], layout2),  
            FigureWidget([], layout3)  
        ]
        self.widgets_cont = VBox([], layout=Layout(width='100%', height="100%"))
        self.debug_output = Output()
        self.qgrid = VBox()
        self.detail = Output()#VBox()

        self.updateMain()
        
    def updateMain( self ):
        self.search = {
            'tool':self.TOOL, 
            'filters':json.dumps([
                {'field':'input.materialp','value':self.materialp.value,'operation':'='},
                {'field':'input.Na','value':self.Na.value/10,'operation':'>='},
                {'field':'input.Na','value':self.Na.value*10,'operation':'<='},
                {'field':'input.Nd','value':self.Nd.value/10,'operation':'>='},
                {'field':'input.Nd','value':self.Nd.value*10,'operation':'<='},    
                {'field':'input.taun','value':self.taun.value/10,'operation':'>='},
                {'field':'input.taun','value':self.taun.value*10,'operation':'<='},    
                {'field':'input.taup','value':self.taup.value/10,'operation':'>='},
                {'field':'input.taup','value':self.taup.value*10,'operation':'<='},    
                {'field':'input.vn_step','value':10,'operation':'>='},    
                {'field':'input.vsweep_high','value':self.vsweep_high.value-0.001,'operation':'>='},    
                {'field':'input.vsweep_high','value':self.vsweep_high.value+0.001,'operation':'<='},    
            ]),
            'results':json.dumps(PNToyExplorer.FILTERS),    
            'revision':PNToyExplorer.REVISION,
            'valid_runs':PNToyExplorer.VALID_RUNS,
            'limit':self.samples,
            'random':PNToyExplorer.RANDOM,
        } 

        result_json = self.session.requestPost('dbexplorer/dbexplorer/stats', data=self.search)
        self.stat_json = result_json.json()
        self.plotWidgets(self.stat_json, PNToyExplorer.FILTERS)        
        self.update_filters()

        
    def codeInput(self, field, values):
        if (field in ['input.taun','input.taup'] and values > 0):
            values = math.log(values,10)
        return values;
    
    def decodeInput(self, field, values):
        if (field in ['input.taun','input.taup']):
            values = math.pow(10, values)
        if (field in ['input.vn_step']):
            values = round(values)
        return values;
            
    def createBands( self ):
        output = ['output.ecx', 'output.eix', 'output.evx']
        traces = []
        for index, row in self.dataframe.iterrows():
            for t in output:
                if row[t] is not None:
                    yaxis = "y"
                    opacity = "0.05"                
                    line = { 'color' : 'rgba(255,0,0,'+opacity+')' }
                    if (t == "output.eix"):
                        line = { 'color' : 'rgba(0,255,0,'+opacity+')' }
                    if (t == "output.evx"):
                        line = { 'color' : 'rgba(0,0,255,'+opacity+')' }

                    trace = {
                      'type': 'scatter',
                      'x':row[t]['xaxis'],
                      'y':row[t]['yaxis'],
                      'yaxis':yaxis,
                      'line' : line,
                      'name' : "SAMPLE " + row['squid'],
                      'hovertemplate':'Length: %{x} Energy: %{y}<br>' + row['squid'] + '<extra></extra>',
                      'showlegend' : False,
                    }
                    self.plotly_fig[0].add_trace(trace)    
                
        for i in range (len(self.plotly_fig[0].data)):
            if (self.plotly_fig[0].data[i].name.startswith("SAMPLE ") ):
                self.plotly_fig[0].data[i].on_click(lambda a, b, c, this=self : this.select_trace(a,b,c))

                
    def updateMS( self ): 
        for data in self.plotly_map[0].data:
            if data.name == "SAMPLE " + self.active:
                data.line.color = 'rgba(255,10,10, 0.8)'
            elif ( data.name.startswith('SAMPLE') ) :
                if (data.line.color != 'rgba(44,162,95, 0.6)'):
                    data.line.color = 'rgba(44,162,95, 0.6)'

    def updateTable (self):
        table = self.qgrid.children[0]
        table.change_selection(rows=self.dataframe.index[self.dataframe['squid']==self.active].tolist())
        
    def updateIV( self ):
        for data in self.plotly_fig[1].data:
            opacity = "0.05"
            if data.name == "SAMPLE " + self.active:
                opacity = "1.0"
            color = 'rgba(255,0,0,'+opacity+')'
            if (data.line.color.startswith('rgba(0,0,255')):
                color = 'rgba(0,0,255,'+opacity+')'
            if data.line.color != color:
                data.line.color = color
            
    def updateBands( self ):       
        xrange = None
        yrange = None
        for data in self.plotly_fig[0].data:
            opacity = "0.05"
            if data.name == "SAMPLE " + self.active:
                opacity = "1.0"
            color = 'rgba(255,0,0,'+opacity+')'
            if (data.line.color.startswith('rgba(0,255,0')):
                color = 'rgba(0,255,0,'+opacity+')'
            if (data.line.color.startswith('rgba(0,0,255')):
                color = 'rgba(0,0,255,'+opacity+')'
            if data.line.color != color :
                data.line.color = color     
                
            if data.name == "SAMPLE " + self.active:
                xr = [min(data.x), max(data.x)]
                yr = [min(data.y), max(data.y)]
                if (xrange is None):
                    xrange = xr
                if (yrange is None):
                    yrange = yr  
                xrange[0] = min(xrange[0],xr[0])
                xrange[1] = max(xrange[1],xr[1])
                yrange[0] = min(yrange[0],yr[0])
                yrange[1] = max(yrange[1],yr[1])
        self.plotly_fig[0].update_xaxes(range=xrange)
        self.plotly_fig[0].update_yaxes(range=yrange)

    def createIV( self ):
        output = ['output.iv', 'output.cap']
        traces = []
        for index, row in self.dataframe.iterrows():
            for t in output:
                if row[t] is not None:
                    yaxis = "y"
                    opacity = "0.05"
                    label = "Current"
                    line = { 'color' : 'rgba(255,0,0,'+opacity+')' }
                    if (t == "output.cap"):
                        yaxis = "y2"
                        line = { 'color' : 'rgba(0,0,255,'+opacity+')' }
                        label = "Capacitance"
                    trace = {
                      'type': 'scatter',
                      'x':row[t]['xaxis'],
                      'y':row[t]['yaxis'],
                      'yaxis':yaxis,
                      'line' : line,
                      'name' : "SAMPLE " + row['squid'],
                      'hovertemplate':'Voltage: %{x} ' + label + ': %{y}<br>' + row['squid'] + '<extra></extra>',
                      'showlegend' : False,
                    }
                    self.plotly_fig[1].add_trace(trace) 
                
        for i in range (len(self.plotly_fig[1].data)):
            if (self.plotly_fig[1].data[i].name.startswith("SAMPLE ") ):
                self.plotly_fig[1].data[i].on_click(lambda a, b, c, this=self : this.select_trace(a,b,c))        

     
        
    def select_trace(self, trace, points, selector):
        with self.debug_output:
            if len(points.point_inds) > 0:
                self.active = points.trace_name.replace("SAMPLE ", "")
                self.session.getRapptureSchema = types.MethodType(lambda s,t,f=True,**k : getRapptureSchema(s, self.TOOL, True, squid=self.active,**k), self.session)
                self.session.getToolParameters = types.MethodType(lambda s, **k : nr.Tools.getToolParameters( s, self.TOOL, **k), self.session)
                with self.detail: 
                    clear_output();
                    print (self.dataframe[self.dataframe['squid'] == self.active])
                self.updateBands()
                self.updateIV()
                self.updateMS()
                self.updateTable()
    
    def plotStats( self, stats, sample, filters ):
        with self.debug_output:
            results = stats['results'][0]
            y=[
                {'name' : '-3STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.2)', 'fillcolor':'rgba(166,189,219, 0.0)', 'hovertext':[]},
                {'name' : '-2STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.4)', 'fillcolor':'rgba(166,189,219, 0.2)', 'hovertext':[]},
                {'name' : '-1STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.6)', 'fillcolor':'rgba(166,189,219, 0.4)','hovertext':[]},
                {'name' : 'MEAN', 'values' :[], 'color':'rgba(166,189,219, 1.0)', 'fillcolor':'rgba(166,189,219, 0.6)', 'mode':'lines', 'hovertext':[]},
                {'name' : '+1STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.6)', 'fillcolor':'rgba(166,189,219, 0.6)', 'hovertext':[]},
                {'name' : '+2STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.4)', 'fillcolor':'rgba(166,189,219, 0.4)', 'hovertext':[]},
                {'name' : '+3STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.2)', 'fillcolor':'rgba(166,189,219, 0.2)', 'hovertext':[]},
            ]
            x=[]
            minlabels = []
            maxlabels = []
            for f in filters:
                if ('MAX_' + f in results and 'MIN_' + f in results):
                    #min_v = self.codeInput(f,results['MIN_' + f])
                    min_v = self.widgets[f].value[0]
                    #max_v = self.codeInput(f,results['MAX_' + f])
                    max_v = self.widgets[f].value[1]
                    avg_v = self.codeInput(f,results['AVG_' + f])
                    if avg_v == None:
                        avg_v = 0
                    std_v = self.codeInput(f,results['STD_' + f])
                    if std_v == None:
                        std_v = 0
                    x.append(f)
                    minlabels.append(min_v)
                    maxlabels.append(max_v)
                    for i, t in enumerate(y):
                        val = avg_v+(i-3)*std_v
                        norm = (max_v-min_v)
                        if norm == 0:
                            norm = 1
                        t['values'].append((val-min_v)/norm)
                        if (t['name'] == "MEAN"):
                            t['hovertext'].append("MEAN (" + str(avg_v) + " +/- " + str(std_v) + ")")

            new_trace = dict(y[0])
            new_trace['fill'] = None
            new_trace['name'] = ""
            y.insert(0,new_trace)

            new_trace = dict(y[0])
            new_trace['values'] = [0 for i in new_trace['values']]
            new_trace['fill'] = None
            new_trace['mode'] = "text"
            new_trace['name'] = "MIN"
            new_trace['hovertext'] = minlabels
            y.append(new_trace)    

            new_trace = dict(y[0])
            new_trace['values'] = [1 for i in new_trace['values']]
            new_trace['fill'] = None
            new_trace['mode'] = "text"
            new_trace['name'] = "MAX"
            new_trace['textposition'] = "bottom center"
            new_trace['hovertext'] = maxlabels
            y.append(new_trace)    

            for t in sample['results']:
                y2 = []
                h2 = []
                for i, v in enumerate(x):
                    norm = (maxlabels[i]-minlabels[i])
                    if norm == 0:
                        norm = 1
                    if (self.codeInput(v, t[v])) == None:
                        y2.append(0.5)
                    else:
                        y2.append((self.codeInput(v, t[v])-minlabels[i])/norm)
                    h2.append(t[v])
                new_trace = dict(y[0])
                new_trace['values'] = y2
                new_trace['fill'] = None
                new_trace['name'] = "SAMPLE " + t['squid']
                new_trace['mode'] = "lines+markers"
                new_trace['color'] = 'rgba(44,162,95, 0.6)'
                new_trace['hovertext'] = h2
                y.append(new_trace)    

            traces = []
            for t in y:
                trace = {
                  'type': 'scatter',
                  'x':[i.replace("input.","") for i in x],
                  'y':t['values'],
                  'hoverinfo' : 'text',
                  'text' : t['hovertext'],
                  'hovertext' : t['hovertext'],
                  'name':t['name'],
                  'mode':t.get('mode', 'none'), 
                  'fill':t.get('fill', 'tonexty'),
                  'fillcolor':t['fillcolor'],
                  'line_shape':'spline',
                  'line_color':t['color'],
                  'textposition':t.get('textposition', 'top center'),
                }
                traces.append(trace)

            self.plotly_map[0].data=[]
            self.plotly_map[0].update({
                'data': traces,
            }) 

            for i in range (len(self.plotly_map[0].data)):
                if (self.plotly_map[0].data[i].name.startswith("SAMPLE ") ):
                    self.plotly_map[0].data[i].on_click(lambda a, b, c, this=self : this.select_trace(a,b,c))

            self.plotly_map[0].layout.xaxis.range = [0,len(x)-1]
            if len(x) > 0:
                dummy_obj = type('', (), {'trace_index':len(traces)-1, 'point_inds':[0]})()
                
    def update_filters( self ):
        with self.debug_output:
            self.active = None
            self.dataframe = None
            self.plotly_fig[0].data=[]
            self.plotly_fig[1].data=[]


            newfilters = [
                {'field':'input.materialp','value':self.materialp.value,'operation':'='},
                {'field':'input.Na','value':self.Na.value/10,'operation':'>='},
                {'field':'input.Na','value':self.Na.value*10,'operation':'<='},
                {'field':'input.Nd','value':self.Nd.value/10,'operation':'>='},
                {'field':'input.Nd','value':self.Nd.value*10,'operation':'<='},  
                {'field':'input.taun','value':self.taun.value/10,'operation':'>='},
                {'field':'input.taun','value':self.taun.value*10,'operation':'<='},    
                {'field':'input.taup','value':self.taup.value/10,'operation':'>='},
                {'field':'input.taup','value':self.taup.value*10,'operation':'<='}, 
                {'field':'input.vn_step','value':10,'operation':'>='},    
                {'field':'input.vsweep_high','value':self.vsweep_high.value-0.001,'operation':'>='},    
                {'field':'input.vsweep_high','value':self.vsweep_high.value+0.001,'operation':'<='},    
            ]
            for k, w in self.widgets.items():
                newfilters.append({'field':k,'value':self.decodeInput(k,w.value[0]),'operation':'>='})
                newfilters.append({'field':k,'value':self.decodeInput(k,w.value[1]),'operation':'<='})

            FILTERS =  [
                'input.Na',
                'input.Nd',
                'input.i_len',
                'input.n_len',
                'input.p_len',
                'input.i_node',
                'input.n_node',
                'input.p_node',
                'input.vn_step',
                'input.materialp',
                'input.temperature',
                'input.vsweep_high',
                'input.impuritylevel',
                'input.impuritydoping'
                'input.taun',
                'input.taup',
                'output.iv', 
                'output.cap',
                'output.ecx', 
                'output.eix', 
                'output.evx'
            ]
            query = {
                'tool':self.TOOL, 
                'filters':json.dumps([]),
                'results':json.dumps(FILTERS),    
                'revision':self.REVISION,
                'valid_runs':self.VALID_RUNS,
                'limit':self.samples,
                'random':PNToyExplorer.RANDOM,
            }
            query['filters'] = json.dumps(newfilters)
            #stat_json = self.session.requestPost('dbexplorer/dbexplorer/stats', data=query)
            result_json = self.session.requestPost('dbexplorer/dbexplorer/search', data=query, timeout=2000)
            result_json = result_json.json()
            self.dataframe = pd.DataFrame(result_json['results'])
            self.qgrid.children = [qgrid.show_grid(self.dataframe)]
            self.plotStats(self.stat_json, result_json, list(self.widgets.keys()))    
            self.createBands()
            self.createIV() 

    def displayWindow(self):   
        tab1 = Tab([
            VBox([
                HBox([VBox([self.materialp, self.vsweep_high, self.samples],layout = Layout(width='50%')), VBox([self.Na,self.Nd,self.taun,self.taup],layout = Layout(width='50%'))], layout=Layout(width='100%', height='150px')),
                HBox([self.plotly_map[1]], layout=Layout(width='100%', height='150px')),
                HBox([self.plotly_map[0]], layout=Layout(width='100%', height='450px')),
            ], layout=Layout(width='100%', height='100%', padding='0px')),
            HBox([self.widgets_cont], layout=Layout(width='100%')),
            HBox([self.qgrid], layout=Layout(width='300%')),
            HBox([self.detail], layout=Layout(width='100%')),
        ],layout=Layout(width='50%'))
        tab1.set_title(0,"Explorer")
        tab1.set_title(1,"Filters")
        tab1.set_title(2,"Dataframe")
        tab1.set_title(3,"Active")
        tab2 = VBox([
            VBox([self.plotly_fig[0]], layout=Layout(width='100%', height='50%', padding='0px')),
            VBox([self.plotly_fig[1]], layout=Layout(width='100%', height='50%', padding='0px')),
        ],layout=Layout(width='50%'))
        self.window.children = [
            HBox([
                tab1,
                tab2
            ], layout=Layout(width='100%', height='640px'))
        ]
        display(self.window)

            
    def plotWidgets( self, stats, filters ):
        with self.debug_output:
            results = stats['results'][0]
            y=[
                {'name' : '-3STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.2)', 'fillcolor':'rgba(166,189,219, 0.0)', 'hovertext':[]},
                {'name' : '-2STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.4)', 'fillcolor':'rgba(166,189,219, 0.2)', 'hovertext':[]},
                {'name' : '-1STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.6)', 'fillcolor':'rgba(166,189,219, 0.4)','hovertext':[]},
                {'name' : 'MEAN', 'values' :[], 'color':'rgba(166,189,219, 1.0)', 'fillcolor':'rgba(166,189,219, 0.6)', 'mode':'lines', 'hovertext':[]},
                {'name' : '+1STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.6)', 'fillcolor':'rgba(166,189,219, 0.6)', 'hovertext':[]},
                {'name' : '+2STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.4)', 'fillcolor':'rgba(166,189,219, 0.4)', 'hovertext':[]},
                {'name' : '+3STDEV', 'values' :[], 'color':'rgba(166,189,219, 0.2)', 'fillcolor':'rgba(166,189,219, 0.2)', 'hovertext':[]},
            ]
            x=[]
            minlabels = []
            maxlabels = []
            avglabels = []
            stdlabels = []
            for f in filters:
                if ('MAX_' + f in results and 'MIN_' + f in results):
                    min_v = self.codeInput(f, results['MIN_' + f])
                    if min_v == None:
                        min_v = 0
                    max_v = self.codeInput(f, results['MAX_' + f])
                    if max_v == None:
                        max_v = 0
                    avg_v = self.codeInput(f, results['AVG_' + f])
                    if avg_v == None:
                        avg_v = 0
                    std_v = self.codeInput(f, results['STD_' + f])
                    if std_v == None:
                        std_v = 0
                    x.append(f)
                    minlabels.append(min_v)
                    maxlabels.append(max_v)
                    avglabels.append(avg_v)
                    stdlabels.append(std_v)
                    for i, t in enumerate(y):
                        val = avg_v+(i-3)*std_v
                        if ( val > max_v ):
                            val = max_v
                        elif ( val < min_v ):
                            val = min_v
                        norm = (max_v-min_v)
                        if norm == 0:
                            norm = 1
                        t['values'].append((val-min_v)/norm)
                        if (t['name'] == "MEAN"):
                            t['hovertext'].append("MEAN (" + str(avg_v) + " +/- " + str(std_v) + ")")

            new_trace = dict(y[0])
            new_trace['fill'] = None
            new_trace['name'] = ""
            y.insert(0,new_trace)

            traces = []
            for t in y:
                trace = {
                  'type': 'scatter',
                  'x':x,
                  'y':t['values'],
                  'hoverinfo' : 'text',
                  'text' : t['hovertext'],
                  'hovertext' : t['hovertext'],
                  'name':t['name'],
                  'mode':t.get('mode', 'none'), 
                  'fill':t.get('fill', 'tonexty'),
                  'fillcolor':t['fillcolor'],
                  'line_shape':'spline',
                  'line_color':t['color'],
                  'textposition':t.get('textposition', 'top center'),
                }
                traces.append(trace)
            tickvals_v = []
            ticktext_v = []
            constraintrange = []
            for i in range(len(x)):
                tickvals = []
                ticktext = []
                for r in range(21):
                    tickvals.append(r/20)
                    ticktext.append(" ".join(["" for i in range(r)]))
                ticktext[0] = minlabels[i]
                ticktext[len(ticktext)-1] = maxlabels[i]
                tickvals_v.append(tickvals)
                ticktext_v.append(ticktext)
                up = avglabels[i]+(0.75*stdlabels[i])
                if (up > maxlabels[i]):
                    up = maxlabels[i]
                down = avglabels[i]-(0.75*stdlabels[i])
                if (down < minlabels[i]):
                    down = minlabels[i]
                norm = (maxlabels[i]-minlabels[i])
                if norm == 0:
                    constraintrange.append([
                        0,
                        1
                    ])
                else:
                    constraintrange.append([
                        ((down)-minlabels[i])/norm,
                        ((up)-minlabels[i])/norm
                    ])
                    
            trace = {
              'type': 'parcoords',
              'tickfont' : {'color' : "black" },
              'dimensions':[{
                  'label':x[i], 
                  'values':[0,1], 
                  'constraintrange':constraintrange[i],
                  'tickvals' : tickvals_v[i],
                  'ticktext' : ticktext_v[i],
                  'multiselect' : False,
              } for i in range(len(x))],
            }
            traces.append(trace)        
            layout = {
            }    

            self.plotly_map[1].data=[]
            self.plotly_map[1].update({
                'data': traces,
                'layout':layout
            }) 
            self.plotly_map[1].layout.xaxis.range = [0,len(x)-1]
            for i in range(len(self.plotly_map[1].data[8].dimensions)):
                self.plotly_map[1].data[8].dimensions[i].on_change(lambda obj, constraintrange, minv=minlabels[i], maxv=maxlabels[i], dim=self.plotly_map[1].data[8].dimensions[i].label, this=self : this.updateValues(dim, constraintrange, minv, maxv), 'constraintrange')

            self.widgets = {}
            children = []

            for i in range(len(x)):
                values = [(constraintrange[i][0]*(maxlabels[i]-minlabels[i]) + minlabels[i]),(constraintrange[i][1]*(maxlabels[i]-minlabels[i]) + minlabels[i])]
                self.widgets[x[i]] = IntRangeSlider( # FloatRangeSlider(
                    value = values, 
                    min=minlabels[i], 
                    max=maxlabels[i], 
                    description=x[i], 
                    orientation='horizontal', 
                    continuous_update = False,
                    style = {'description_width': 'initial'},
                    layout = Layout(width='98%')
                )
                self.widgets[x[i]].observe(lambda c : self.updateFilters(), names='value')
                children.append(self.widgets[x[i]])
            children.append(self.materialp)
            children.append(self.vsweep_high)
            children.append(self.samples)
            children.append(self.Na)
            children.append(self.Nd)
            children.append(self.taun)
            children.append(self.taup)
            self.widgets_cont.children = children
            
    def updateFilters(self):
        with self.debug_output:
            for i,dim in enumerate(self.plotly_map[1].data[8].dimensions):
                vmax = self.widgets[dim.label].max
                vmin = self.widgets[dim.label].min
                if (vmax == vmin):
                    vmax = vmax+1
                values = [(val-vmin)/(vmax-vmin) for val in self.widgets[dim.label].value]
                dim.constraintrange = values
            self.update_filters()
        
    def updateValues(self, dimension, constraintrange, minv, maxv):
        if constraintrange == None:
            values = [minv,maxv]
        else:
            values = [(constraintrange[0]*(maxv-minv) + minv),(constraintrange[1]*(maxv-minv) + minv)]

        self.widgets[dimension].value = values